package Core;

use strict;
use warnings;

=head1 Core.pm

Common functions and data for reports.

Handles combined data with various keys grouping and aggregation
functions.  Generates localized output and translated messages.

=head1 Remarks

TODO handle selections with filtering expressions;
     should work like "select: blablabla for EXPR"

TODO handle selections with functions included
     separately for given group, and with function
     with associated BY-grouping.

TODO Add support for Graph generation.

=head2 Variable naming mnemonics

To deconfuse myself and anyone reading the code.

 field		is a part of log record;

 label		is an (internal) name of field,
		lowercased and w/out spaces;

 title		printable and translated label of keyfield
		or aggregation function - to use in tables;

 aggregation	function used to aggregate values of some field
		for set of record defined by grouping;

 keyfields	fields used to index data, that is - fields of
		countable data;

 group		is a list of keyfields defined in config to
		describe how aggregations will be performed in a
		group fields go in the order given in config to
		define AND an order of data appearence in tables;

 index		is an ordered list of keyfields, used to access
		data; in an index fields go in the internal order of
		data module to do not use different indexes for the
		same data; thus - the same index may be used for
		several groups, containing the same set of
		keyfields;

 key		is an association of all keyfields in index with
		some keyvalues, used to actual access data tree;
		when it's a hash, order of fields is destroyed
		and is to be restored by Data::ptr;

 selection	set of groups defined in single directive 'select'
		using grouping expression or just comma-separated
		list of keyfields, may contain separated subsets;

=cut

use vars
qw#	$extra @keyfields @selections @aggregations %aggregations
	%datatypes %datatitles @sortorder $reportype_title @pages
	$data_ref $table_refname $table_caption#;
use locale; # for sprintf
use POSIX qw#strftime#;
use Exporter;
use Carp;
use Config qw#get_config#;
use Data;
use Setup qw#datalabels datatypes selections#;;
use Utils qw#filescape trans#;
use Html qw#foot head escape#;
use Format qw#comp i18n_format#;
use Trans qw#translate#;

@Core::ISA = qw#Exporter#;

@Core::EXPORT_OK =
qw#	$extra @keyfields @selections @aggregations %aggregations
	%datatypes %datatitles @sortorder $reportype_title @pages
	$data_ref $table_refname $table_caption
	
	keysort sortindex flatkey flatkeys amount unroll setup
	format_titles strfval file_idx_ref page_idx_title generate_table
	generate_table_f generate_general generate_head#;

*extra = *::EXTRA;

$Core::NAME = 'Core';
$Core::VERSION = '0.1.0';
$Core::AUTHOR = 'qMax';
$Core::EMAIL = 'qmax@mail.ru';

=head1 Functions and data

=cut

=head2 @keyfields

List of selected fields. Names are lowercased.

=cut

@keyfields=();

=head2 @selections

Contains list of group selections.
Each element is a list of sets of groups selected in single
directive. Selection may contain several separated subsets.

 For single group:	[ [ group ] ]

 For set of groups:	[ [ group group ] ]

 For separated groups:	[ [ group group ] [ group group ] ]

 group is itself [keyfield, keyfield, ...]

All groups are scanned to figure out @keyfields.

=cut

@selections=();

=head2 @aggregations %aggregations

@aggregations is a list of selected aggregations in order, given
in config.  Names of aggregations are in form "FUN(arg)".

%aggregations is definition of aggregations in form:

    ( name=>[$fun,$arg] )

The [$fun,$arg] is a part of key to look up data with Data::ptr()

Possible hack is to refer aggregations by value of reference to
[$fun $fld] instead of "$FUN($arg)" string, then %aggregations
will be unneeded. (values %aggr) don't preserve order, and
should not be used instead of @aggr.

=cut

@aggregations=();
%aggregations=();

=head2 %datatypes

Types of data to properly format and sort,
including types for aggregation fields,
either keyfields or aggregations.

    (name => $type)

=cut

%datatypes=();

=head2 %datatitles

Printable and translated labels of data fields to mark table cells,
either keyfields or aggregations.

=cut

%datatitles=();

=head2 $reportype_title

Messages to put in html title to denote report type.
Should be setup in report module.

=cut

$reportype_title="";

=head2 $data_ref $table_refname $table_caption

Reference to subs used to table captions and 
crossreferences from data cells to other tables.
Used in generate_table* subs.

If $table_refname and $table_caption are defined they are used
as &$table_foo(%BASEKEY) to insert table caption and 'A name' tag.
If there's single aggregation in table passed, it is included 
in the %KEY ().

If $data_ref is defined it is used as &$dataref(%KEY) to insert
'A href' tag in table cells. %KEY may include aggregation -
then it points to data cell.
Filtering ignore_less is then skipped to do not miss any references.

=cut

$data_ref=undef;
$table_refname=undef;
$table_caption=undef;

=head2 @pages

Should be set to contains list of categories generated by report
module.

Elements are passed to page_title() and file_ref(), defined in
report module.

Used from generate_head().

=cut

@pages=();

=head2 keysort

To use as sort sub for lists keys. Orders keys such that keyvalues
of sequential keys get sorted according to their types.
Priority of keynames should be set up in @sortorder.
Compares elements with i18n::sort() according to types of keynames.

Larry promised, such sub is inlined.

=cut

@sortorder=();

print "\nCore.pm (Core) loaded. Returning." if ($::DEBUG & 0x01);

return (1);

# what a nice bullshit - this is to reverse export ::a and ::b
sub keysort
{
	my $l = $Report::a;
	my $r = $Report::b;
	
	foreach (@sortorder)
	{
		if ($l->{$_} ne $r->{$_})
		{
			return comp ($datatypes{$_}, $l->{$_}, $r->{$_});
		}
	}
	return 0;
}

=head2 sortindex @FIELDS

Sorts fields in order of @sortorder

=cut

sub sortindex {
    my @i=@_;
    my @r=();
    foreach my $kf (@sortorder) { push @r,$kf if( grep($kf eq $_,@i) );	}
    return @r;
}

=head2 flatkey %KEY

Coerces %KEY into even-sized array with order of keynames as in @sortorder.

=cut

sub flatkey
{
    my $key=shift;
    my @flat=();
    foreach (sortindex(keys %$key)) { push @flat,($_,$key->{$_}); }
    return @flat;
}

sub flatkeys
{
    my $keys=shift;
    my @newkeys=();
    map{ push @newkeys,[flatkey($_)] } @$keys;
    return @newkeys;
}

=head2 setup

Setups all global variables.
Does not initialize data module to allow derived modules to add something.
Does not check if selected keyfields and functions are correct -
this is report- and data/parser- dependant.

=cut

# experimental
use vars qw(@assocs);
@assocs=();

sub setup
{
    my @datalabels = datalabels();
    %datatypes = datatypes();

    # titles should be defined prior to aggregation scanning
    foreach (@datalabels) {
	$datatitles{$_}=translate($_);
	}

    # scan selections
    my @jopa=selections();
    foreach my $sel (@jopa) {
	$_ = shift @$sel;
	if( /FUN/ ) {
	    my ($fun,$fld)= @$sel;
	    my $aggr="$fun($fld)";
	    push @aggregations,$aggr;
	    $aggregations{$aggr}=[$fun,$fld];

	    if( $fun eq 'COUNT' ) {
		$datatypes{$aggr}='number'; 
		if( $fld eq '*' ) { $fld='record'; }
		$datatitles{$aggr} = translate("$fun(%s)",$fld);
		}
	    else {
		$datatypes{$aggr}=$datatypes{$fld};
		$datatitles{$aggr} = translate("$fun(%S)",$fld);
		}
	    }
	    # experimental code
	elsif( /SEL/ ) {
	    my ($fun,$fld,$ongr,$bygr)= @$sel;
	    push @assocs, [ $ongr, [[$fun, $fld]], $bygr ];
	    }
	elsif( /GRP/ ) {
	    my @groupsets = @$sel;
	    push @selections,[ @groupsets ];
	    # scan keyfields
	    foreach my $gset (@groupsets) {
		foreach my $grp (@$gset) {
		    foreach my $fld (@$grp) {
			push @keyfields, $fld unless( grep($fld eq $_,@keyfields ) );
			}
		    }
		}
	    }
	# ignore unknown tags
	}
}

=head2 amount ($key,$aggregation)

Retrieves data from pool, the $aggregation is internal name of aggr.
Returns 0 if specified key does not exists.

=cut

sub amount 
{
    my ($k,$f)=@_;
    my $dataref=Data::exists({%$k,@{$aggregations{$f}}});
    return ($dataref ? $$dataref : 0);
}

=head2 unroll \@fields \%basekey

Returns keys of all avaliable keyvalues of fields in @fields 
under given %basekey.
Basekey is not included in returning keys.

=cut

sub unroll
{
    my ($fields,$basekey)=@_;
    if( $basekey ) {
	my @keys = Data::grep( %$basekey, map{ $_=>undef } @$fields );
	return map{ my $k=$_; map{delete $k->{$_}} keys %$basekey; $k; } @keys;
	}
    else {
	return Data::grep( map{ $_=>undef } @$fields );
	}
}

=head2 format_period ($start, $end)

Returns strftime formatted reporting period in form
"FooStart - FooEnd Bar", where Bar is a least common part of
dates, and Foo are greatest different parts of start and end
dates. Looks nice.

TODONOT: calculate a seasons for periods across several months within
	 same year, or across boundaries.

=cut

sub format_period
{
    my($start,$end)=@_;
    return undef unless($start and $end);
    my @fmtdiff=('%H:%M:%S','%H:%M','%H:%M','%d','%d %B','%B %Y');
    my @fmtcomm=('%d %B',   '%d %B','%d %B','%B','%Y',   'a.c.');
    my @start = localtime($start);
    my @end = localtime($end);
    my ($i);
    for($i=5; $i>=0 && ($start[$i] eq $end[$i]);$i--) { }
    return
	POSIX::strftime($fmtdiff[$i],@start).
	" - ".
	POSIX::strftime($fmtdiff[$i],@end).
	" ".
	POSIX::strftime($fmtcomm[$i],@start);

}

=head2 format_titles ()

Prepares common report title and header for all pages.
Should be callen after data is accumulated.
Used in generate_head.

Otherwise, $report_title and $report_header should be set.

=cut

use vars qw($report_title $report_header);
$report_title="";
$report_header="";

sub format_titles
{
	my $logtype = get_config ('logtype');
    $reportype_title = $Report::VERSION unless ($reportype_title);

    my @title;
    push @title, $::NAME . ' ' . $::VERSION;
    push @title, "$logtype / $reportype_title";

    if( $Parser::extra{'start'} and $Parser::extra{'end'} ) {
	my $period_title = format_period($Parser::extra{'start'},
			      		 $Parser::extra{'end'});
	push @title, $period_title;
	$report_header = translate("Report for")." $period_title";
	}
    else {
	$report_header = "$logtype / $reportype_title";
	}

    $report_title  = join(' : ',@title);
}

=head2 strfval ($key,$val)

Returns $val formated according to type of $key

=cut

sub strfval 
{
    my ($key, $val) = @_;
    return (escape (i18n_format ($datatypes{$key}, $val)))[0];
}

=head2 print_precentage ($precentage)

Returns properly formatted <td>precentage<td>bars.
The $precentage is in range 0..1

=cut

sub print_precentage
{
	my $prec = shift;
	my $text;

	my $cfg_width  = get_config ('bar_width' );
	my $cfg_height = get_config ('bar_height');
	my $cfg_img_fg = get_config ('bar_image_foreground');
	my $cfg_img_bg = get_config ('bar_image_background');
	
	$cfg_width  ||= '64';
	$cfg_height ||= '8';
	$cfg_img_bg ||= 'dot-dark.png';
	$cfg_img_fg ||= 'dot-light.png';
	
	$text = sprintf ("\n    <td>%.2f%%</td>", $prec * 100);

	my $width = int ($cfg_width * $prec);
	$width = $cfg_width if ($width > $cfg_width);
	
	$text .= "\n    <td>";
	
	$text .= qq#<img src="$cfg_img_fg" width="$width" height="# .
		$cfg_height . '" />'
		if ($width);

	if ($cfg_img_bg and $cfg_width > $width)
	{
		$width = $cfg_width - $width;
		$text .= qq#<img src="$cfg_img_bg" width="$width" height="# .
			$cfg_height . '" />';
	}

	$text .= '</td>';

	return $text;
}

=head2 filter (\@keys, \@ignored, \%totals)

Filters list of keys against config->{ignore_less}
Returns number of ignored records (keys),
updates @keys and puts ignored keys into @ignored.

The %totals is ( aggregation => totalfortable ) to use for calculate %%.

TODO: This is in general a place to do WHERE filtering.
      Thus, some definition of filtering criteria
      should be passed as an arg, instead of totals.
      Filtering criteria may be an or-list of and-lists
      of simple expressions, or, simpler - a perl expr to eval. 
      (this will lack of IN, ANY, EVERY statments).

TODO: A good idea is to precalculate precentages before filtering
      and temporary store them in data hash. This will generalize
      filtering.

=cut

sub filter
{
    my ($basekey,$keys,$ignored,$totals) = @_;
    my $ignoring= get_config ('ignore');
    my @filtered=();
    @$ignored=();

    if ($ignoring and ($ignoring >= 0 and $ignoring <= 100))
    {
	    $ignoring /= 100;
    }
    else
    {
	print STDERR $/, __FILE__, ": Ignore value '$ignoring' invalid. ",
		'Assuming 1  percent.' if ($ignoring);
    	$ignoring = 0.01;
    }
    
    # there possible several aggregates in a table,
    # thus using maximal %% to check
    foreach my $key (@$keys) {
	my $maxprec=0;
	my %cur_amnt=();
	foreach (keys %$totals) {
	    # AVG is not a subject to filter
	    next if( /AVG/ );
	    my $amount = amount({%$basekey,%$key},$_);
	    my $prec = $totals->{$_} ? $amount/$totals->{$_} : 0;
	    $maxprec=$prec if($prec> $maxprec);
	    $cur_amnt{$_}=$amount;
	    }
	if( $maxprec < $ignoring ) {
	    push @$ignored,$key;
	    }
	else {
	    push @filtered,$key;
	    }
	}

    @$keys = @filtered if( @filtered );

    return scalar(@$ignored);
}

#
# row/col spanning algo:
# each skiped cell decreases spancount,
# new cell is printed when it reaches 0
# and then spancount recalculated.
#

# count matching keys from $ki to first nonmatching
# this will correctly rowspan if keys are in random order.
# NB: $key doesn't have all keyfields as in $keylist
sub countspans
{
    my ($key,$keylist,$ki)=@_;
    my $len = scalar(@$keylist);
    my $cnt=0;

    while($ki < $len) {
	last if( grep($key->{$_} ne $keylist->[$ki]{$_},keys %$key) );
	$cnt++;
	$ki++;
	}

    return $cnt;
}

=head2 generate_table $FILE,\%basekey,\@aggregations,\@rowkeys,\@rowindex,\@colkeys,\@colindex

Generates single table. Creates caption and crossreferences.
Does not check if data under some keys actually exists.

 %basekey is common key, appended to all other keys;
 @aggregations is list of aggregations to include;
 @rowkeys is a sorted list of keys for rows;
 @rowindex is order of row fields;
 @colkeys is a sorted list of keys for columns;
 @colindex is order of col fields;

TODO: passing fieldorder is a kludge, and forces to pass aggregations
      separately. But hash assignments destroy key order.
      Possible solution is to use even-sized arrays instead
      of hashes. this will also speed up iterations and
      allow to include aggregations within (any) keys.

TODO: print total amounts for subgroups, the same way as it is 
      done when there're several aggregations in 2D table.

=cut

sub generate_table
{
    my ($FILE,$basekey,$aggregations,$rows,$rowindex,$cols,$colindex)=@_;
    my @aggregations=@$aggregations;
    my @rowkeys=@$rows;
    my @rowindex=@$rowindex;
    my $rowlen=scalar(@rowindex);
    my $rowscnt=scalar(@rowkeys);
    my @colkeys=();
    my @colindex=();
    my $collen=0;
    my $colscnt=0;
    if( $cols ) {
	@colkeys=@$cols;
	@colindex=@$colindex;
	$collen=scalar(@colindex);
	$colscnt=scalar(@colkeys);
	}

	print $FILE qq#\n\n<div><table>#;

    my ($caption,$refname)=('','');
    my %refkey=();
    if( scalar(@aggregations) > 1 ) { %refkey=%$basekey; }
    else { %refkey=( %$basekey, @{$aggregations{$aggregations[0]}} ); }
    $caption = &$table_caption(%refkey) if( defined $table_caption );
    $refname = &$table_refname(%refkey) if( defined $table_refname );

	if ($refname)
	{
		print $FILE qq#\n<caption><a name="$refname">$caption</a></caption>#;
	}
	else
	{
		print $FILE "\n<caption>$caption</caption>";
	}

    # total amounts for whole table 
    # used to calculate %% and gnore_less filtering
    my %totals=map { $_=>amount($basekey,$_) } @aggregations;

    # filter out keys. this IS a kludge. see Remarks.
    my @ignored_rows=();
    if( not $data_ref and scalar(keys %totals) ) {
	filter($basekey, \@rowkeys, \@ignored_rows, \%totals);
	if( scalar(@ignored_rows) == 1 or scalar(@rowkeys)==0) {
	    push @rowkeys, @ignored_rows;
	    @ignored_rows=();
	    }
	$rowscnt=scalar(@rowkeys);
	}

    my $rowspan_text='';
    my $colspan_text='';

    # COLS headers

	my @colheaders=(); # contains rows of headers
	my %colspans; 
	foreach my $kf (@colindex)
	{
		my $text="";
		my %curkey=();
		foreach my $ki (0 .. $colscnt-1)
		{
			my $kv=$colkeys[$ki]->{$kf};
			$curkey{$kf}=$kv;
			if( --$colspans{$kf} < 0)
			{
				$colspan_text='';
				$colspans{$kf}=countspans(\%curkey,\@colkeys,$ki);
				$colspan_text = qq# colspan="$colspans{$kf}"# unless($colspans{$kf}==0);
				$colspans{$kf}--;
				$text .= qq#\n    <th class="subhdr"$colspan_text>#;
				my $html_text=strfval($kf,$kv);
				my $link; $link=&$data_ref(%curkey) if( $data_ref );
				if( $link ) { $text .= qq#<a href="$link">$html_text</a>#; }
				else        { $text .=                  $html_text; }
				$text .= '</th>';
			}
		}
		push @colheaders, $text;
	}

	print $FILE $/, '<!-- ', __FILE__, ' [', __LINE__, '] -->';
	print {$FILE} "\n  <tr>";

	$rowspan_text = "";
	$rowspan_text = sprintf ('rowspan="%u"', $collen + 1);
	$colspan_text = sprintf ('colspan="%u"', $rowlen);

	print $FILE qq#\n    <td id="logo" $rowspan_text $colspan_text><img src="logo.png" /></td>#;

	my $have_rowtotals='';
	if( $colscnt > 0 )
	{
		$rowspan_text = sprintf ('rowspan="%u"', $collen);
		$have_rowtotals .= qq#\n    <th class="subhdr" colspan="3" $rowspan_text>#;
		$have_rowtotals .= translate ('Total for %s', $rowindex[$#rowindex]);
		$have_rowtotals .= '</th>';
	
	}

	$colspan_text = sprintf ('colspan="%u"', $colscnt + 3);
	foreach my $aggr ( @aggregations )
	{
		print $FILE "\n    <th $colspan_text>", $datatitles{$aggr}, '</th>';
		print $FILE '<!-- ', __FILE__, ' [', __LINE__, '] -->';
	}
	print $FILE "\n  </tr>";

	if( my $h = shift @colheaders ) {
		print $FILE "\n  <tr>";
		print $FILE "$h$have_rowtotals" x scalar(@aggregations);
		print $FILE "\n  </tr>";
		print $FILE '<!-- ', __FILE__, ' [', __LINE__, '] -->';
	}
	
	while( my $h = shift @colheaders)
	{
		print $FILE "\n  <tr>";
		print $FILE "$h" x scalar(@aggregations);
		print $FILE "\n  </tr>";
		print $FILE '<!-- ', __FILE__, ' [', __LINE__, '] -->';
	}

	# ROWS
	my %rowspans = ();
	foreach my $ki (0 ... ($rowscnt - 1))
	{
		my $rkey = $rowkeys[$ki];
		print $FILE "\n  <tr>";

		# header
		my %curkey = ();
		# kf == keyfield
		foreach my $kf (@rowindex)
		{
			# kv == key value
			my $kv = $rkey->{$kf};
			$curkey{$kf} = $kv;
			if (--$rowspans{$kf} < 0)
			{
				my $rowspan = '';
				$rowspans{$kf} = countspans (\%curkey, \@rowkeys, $ki);
				$rowspan = ' rowspan="' . $rowspans{$kf} . '"'
					unless ($rowspans{$kf} == 0);
				$rowspans{$kf}--;
				
				my $html_text = strfval ($kf, $kv);
				my $link;
				$link = &$data_ref (%curkey) if ($data_ref);

				print $FILE qq#\n    <th class="subhdr"$rowspan>#;
				if ($link) { print $FILE qq#<a href="$link">$html_text</a>#; }
				else { print $FILE $html_text; }
				print $FILE '</th>';
			}
		}

		# data
		foreach my $aggr (@aggregations)
		{
			if ($colscnt)
			{
				foreach my $ckey (@colkeys)
				{
					my %curkey = (%$basekey, %$rkey, %$ckey);
					my $amount = amount (\%curkey, $aggr);
					my $html_text = $amount ? strfval ($aggr, $amount) : '-';
					my $link;
					$link = &$data_ref (%curkey, @{$aggregations{$aggr}})
						if ($data_ref);
					
					print $FILE "\n    <td>";
					if ($link) { print $FILE qq#<a href="$link">$html_text</a>#; }
					else { print $FILE $html_text; }
					print $FILE '</td>';
				}
			}
			
			# total for $aggr
			my %curkey = (%$basekey,%$rkey);
			my $amount = amount (\%curkey, $aggr);
			my $total = $totals{$aggr};
			my $prec = $total ? ($amount / $total) : 0;
			my $html_text = strfval ($aggr, $amount);
			my $link;
			$link = &$data_ref (%curkey, @{$aggregations{$aggr}})
				if ($data_ref);
			
			print $FILE "\n    <td>";
			if ($link) { print $FILE qq#<a href="$link">$html_text</a>#; }
			else { print $FILE $html_text; }
			print $FILE '</td>';
			print $FILE print_precentage ($prec);
		}
		
		print $FILE "\n  </tr>";
	}

	if (@ignored_rows)
	{
		my %ignored_totals=();

		print $FILE "\n  <tr>";
		print $FILE qq#\n    <th class="subhdr" colspan="# . $rowlen . '">',
			translate ('More %d skipped', scalar @ignored_rows),
			"</th>";
	
		foreach my $aggr (@aggregations)
		{
			if ($colscnt)
			{
				# accumulate totals for ignored rows on each $ckey
				foreach my $ckey (@colkeys)
				{
					my $amount=0;
					foreach my $rkey (@ignored_rows)
					{
						$amount += amount({%$basekey,%$rkey,%$ckey},$aggr);
					}
					print $FILE "\n    <td>" . strfval ($aggr, $amount) .
						'</td>';
				}
			}

			# get totals for ignored rows on $aggr
			my $amount = 0;
			my $prec = 0;
			foreach my $rkey (@ignored_rows)
			{
				$amount += amount({%$basekey,%$rkey},$aggr);
				$prec   += $totals{$aggr} ? $amount/$totals{$aggr} : 0;
			}

			print $FILE "\n    <td>" . strfval ($aggr, $amount) . '</td>';
			print $FILE print_precentage ($prec);
		}

		print $FILE "\n  </tr>";
	}

	# COLS: totals
	print $FILE "\n  <tr>";
	print $FILE qq#\n    <th class="subhdr" colspan="#
		. $rowlen . '">';

	if ($colscnt)
	{
		print $FILE translate ('Total for %s', $colindex[$#colindex]),
			'</th>';

		foreach my $aggr (@aggregations)
		{
			foreach my $ckey (@colkeys)
			{
				my $amount = amount ({%$basekey, %$ckey}, $aggr);
				my $total=$totals{$aggr};
				my $prec= $total ? $amount / $total : 0;
				
				print $FILE "\n    <td>", strfval ($aggr, $amount);
				my $html_text = print_precentage ($prec);
				$html_text =~ s#<td>#<br />#sg;
				$html_text =~ s#</td>##sg;
				$html_text =~ s#\n|\s{4,}##sg;
				print $FILE $html_text, '</td>';
			}
		
			# grand total for $aggr
			my $amount = amount ({%$basekey}, $aggr);

			print $FILE "\n    <td>", strfval ($aggr, $totals{$aggr}),
				'<br />&nbsp;<br />&nbsp;</td>';

			my $html_text = print_precentage(1);
			$html_text =~ s#<td>#<td>&nbsp;<br />#s;
			$html_text =~ s#</td>#<br />&nbsp;</td>#s;
			$html_text =~ s#<img#&nbsp;<br />&nbsp;<br /><img#s;
			print $FILE $html_text;
		}
	}
	else
	{
		# total for basekey
		print $FILE translate('Total for %s','Reporting period'),
			'</th>';
		foreach my $aggr (@aggregations)
		{
			print $FILE "\n    <td>" . strfval ($aggr, $totals{$aggr}) . '</td>';
			print {$FILE} print_precentage(1);
		}
	}
	
	print $FILE "\n  </tr>\n</table></div>";
}

=head2 generate_general ($FILE)

Generates general page. Includes:
    Pretty formated period from %Parser::extra{begin,end}
    Total amount for all aggregations (requires empty index to be used)
    Extra info, if any.

=cut

sub generate_general
{
	my ($FILE) = @_;

	generate_head ($FILE);

	print $FILE "\n<table>";

	if ($Parser::extra{'start'} and $Parser::extra{'end'})
	{
		my $duration = int (($Parser::extra{'end'} - $Parser::extra{'start'}) / (24 * 60 * 60));

		print $FILE qq#\n  <tr>\n    <th colspan="2">#,
			translate('Reporting period'),
			"</th>\n  </tr>";

		print $FILE qq#\n  <tr>\n    <th class="subhdr">#,
			translate ('Begin'), '</th>';
		print $FILE "\n    <td>",
			i18n_format ('datetime', $Parser::extra{'start'}),
			"</td>\n  </tr>";

		print $FILE qq#\n  <tr>\n    <th class="subhdr">#,
			translate ('End'), '</th>';
		print $FILE "\n    <td>",
			i18n_format ('datetime', $Parser::extra{'end'}),
			"</td>\n  </tr>";
		
		print $FILE qq#\n  <tr>\n    <th class="subhdr">#,
			translate ('Total of days'), '</th>';
		print $FILE "\n  <td>$duration</td>\n  </tr>";
	}

	print $FILE "\n  <tr><th colspan=2>" . translate('Total') . "</th></tr>";

	foreach my $agg (@aggregations)
	{
		if (Data::exists ({@{$aggregations{$agg}}}))
		{
			print $FILE qq#\n  <tr>\n    <th class="subhdr">#,
				$datatitles{$agg}, "</th>";
			print $FILE "\n    <td>";
			print $FILE i18n_format ($datatypes{$agg}, amount ({}, $agg));
			print $FILE "</td>";
			print $FILE '<!-- ', __FILE__, ' [', __LINE__, '] -->';
		}
	}

	if (%$extra)
	{
		print $FILE qq#\n  <tr><th colspan="2">#,
			translate('Extra'), "</th></tr>";

		while (my ($key, $val) = each %$extra)
		{
			next if ($key eq 'period');

			print $FILE qq#\n  <tr>\n    <th class="subhdr">#,
				trans($key), "</th>";	
			print $FILE "\n    <td>$val</td>\n  </tr>";

			print $FILE '<!-- ', __FILE__, ' [', __LINE__, '] -->';
		}
	}

	print $FILE "\n</table>\n";

	print {$FILE} foot();
}

=head2 file_idx_ref (@INDEX) page_idx_title (@INDEX)

Subs to help composing file names and page titles.
The @INDEX is list of key fields.

 file_idx_ref returns "foo-bar-...etc" w/out .html
 page_idx_title returns "Foos, Bars, etc" with 's' added

NB: Order of keys in INDEX is not changed.

=cut

sub file_idx_ref {
    my @idx=@_;
    return join("-", map(filescape($_), @idx) );
}

sub page_idx_title {
    my @idx=@_;
    return trans('/%s/, /',map($_.'s',@idx));
}

=head2 generate_head ($FILE,@category)

Generates common top-index, including 'General' and @pages.
@category expected to be one of elements in @pages.
Empty category means 'General'.

Since @category is report-dependant, there MUST be defined:

 report::page_title(@category)  - to print page title used in HTML title
                                  and top-page index;
 report::page_header(@category) - to print full page header;
 report::file_ref(@category)    - to make references from topindex;

They should know what actually contained in @category and @groups
and how to print it.

=cut

sub generate_head
{
	# Needs to be imported at runtime..  -octo
	import Report qw#file_ref page_title page_header#;
	
	my ($FILE,@category) = @_;
	my ($link,$title,$page_title,$page_header);

	$page_title =  ( @category ) ? page_title(@category) : trans('General');
	$page_header = ( @category ) ? page_header(@category) : trans('General');

	print $FILE head ("$report_title : $page_title", $report_header);

	print $FILE qq#\n<div id="index">#;

	$title=trans ('General');
	if (@category)
	{
		print $FILE qq#\n  <a class="index" href="index.html"># .
			"[\&nbsp;$title\&nbsp;]</a>";
	}
	else
	{
		print $FILE "\n  [\&nbsp;$title\&nbsp;]";
	}

	# checking for current category is complicated
	# when using references, arrays and undefs
	# and anyway makes not much sense
	foreach my $cat (@pages)
	{
		$link = file_ref(@$cat);
		$title = page_title(@$cat);

		print $FILE qq#\n  <a class="index" href="$link"># .
			"[\&nbsp;$title\&nbsp;]</a>";
	}
	print $FILE "\n</div>\n";
	print $FILE "\n<hr />\n";

	print $FILE "\n<h2>$page_header</h2>";
}

=head1 AUTHOR

QMax E<lt>qmax@mail.ruE<gt>

=cut
